@@ -13,57 +13,93 @@ package LaTeXML::Common::Error;
 use strict;
 use warnings;
 use LaTeXML::Global;
-##use LaTeXML::Common::Object;
+use LaTeXML::Common::Object;
 use Time::HiRes;
+use Term::ANSIColor 2.01 qw(colored colorstrip);
+
 use base qw(Exporter);
 our @EXPORT = (
   # Error Reporting
   qw(&Fatal &Error &Warn &Info),
   # Progress reporting
-  qw( &NoteProgress &NoteProgressDetailed &NoteBegin &NoteEnd),
+  qw(&NoteProgress &NoteProgressDetailed &NoteBegin &NoteEnd),
+  # Colored-logging related functions
+  qw(&colorizeString),
+  # stateless message generation
+  qw(&generateMessage),
+  # Status management
+  qw(&MergeStatus),
 );
 
 #%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
-# Note: The exported symbols should ultimately be exported as part
-# of LaTeXML::Common, or something like that, to be used BOTH in
-# Digestion & Post-Processing.
-# ======================================================================
-# We want LaTeXML::Global to import this package,
-# but we also want to use some of it's low-level functions.
-sub ToString {
-  my ($item, @more) = @_;
-  return ($LaTeXML::BAILOUT ? "$item" : LaTeXML::Common::Object::ToString($item, @more)); }
+# Color setup
+$Term::ANSIColor::AUTORESET = 1;
+our $COLORIZED_LOGGING = -t STDERR;
 
-sub Stringify {
-  my ($item, @more) = @_;
-  return ($LaTeXML::BAILOUT ? "$item" : LaTeXML::Common::Object::Stringify($item, @more)); }
+our %color_scheme = (
+  details => 'bold',
+  success => 'green',
+  info    => 'bright_blue',          # bright only recently defined
+  warning => 'yellow',
+  error   => 'bold red',
+  fatal   => 'bold red underline',
+);
 
+sub colorizeString {
+  my ($string, $alias) = @_;
+  return ($COLORIZED_LOGGING && $color_scheme{$alias}
+    ? colored($string, $color_scheme{$alias})
+    : $string); }
 #%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 # Error reporting
 # Public API
 
 sub Fatal {
   my ($category, $object, $where, $message, @details) = @_;
-  my $state = $STATE;
+
+# Check if this is a known unsafe fatal and flag it if so (so that we reinitialize in daemon contexts)
+  if ((($category eq 'internal') && ($object eq '<recursion>')) ||
+    ($category eq 'too_many_errors')) {
+    $LaTeXML::UNSAFE_FATAL = 1; }
+
+  # We'll assume that if the DIE handler is bound (presumably to this function)
+  # we're in the outermost call to Fatal; we'll clear the handler so that we don't nest calls.
+  die $message if $LaTeXML::IGNORE_ERRORS    # Short circuit, w/no formatting, if in probing eval
+    || (($SIG{__DIE__} eq 'DEFAULT') && $^S);    # Also missing class when parsing bindings(?!?!)
+
+  # print STDERR "\nHANDLING FATAL:"
+  #   ." ignore=".($LaTeXML::IGNORE_ERRORS || '<no>')
+  #   ." handler=".($SIG{__DIE__}||'<none>')
+  #   ." parsing=".($^S||'<no>')
+  #   ."\n";
+  my $inhandler = !$SIG{__DIE__};
+  my $ineval    = 0;                # whether we're in an eval should no longer matter!
+
+  # This seemingly should be "local", but that doesn't seem to help with timeout/alarm/term?
+  # It should be safe so long as the caller has bound it and rebinds it if necessary.
+  local $SIG{__DIE__} = 'DEFAULT';    # Avoid recursion while preparing the message.
+  my $state     = $STATE;
   my $verbosity = $state && $state->lookupValue('VERBOSITY') || 0;
-  if (!$LaTeXML::Common::Error::InHandler && defined($^S)) {
+
+  if (!$inhandler) {
     local $LaTeXML::BAILOUT = $LaTeXML::BAILOUT;
     if (checkRecursiveError()) {
       $LaTeXML::BAILOUT = 1;
       push(@details, "Recursive Error!"); }
-    $state && $state->noteStatus('fatal');
+    $state->noteStatus('fatal') if $state && !$ineval;
+    my $detail_level = (($verbosity <= 1) && ($category =~ /^(?:timeout|too_many_errors)$/)) ? 0 : 2;
     $message
-      = generateMessage("Fatal:" . $category . ":" . ToString($object), $where, $message, 1,
-      # ?!?!?!?!?!
-      # or just verbosity code >>>1 ???
-      @details,
-      ($verbosity > 0 ? ("Stack Trace:", stacktrace()) : ()));
-    # We're about to DIE, which will bypass the usual status message, so add it here.
-    $message .= $state->getStatusMessage if $state;
+      = generateMessage(colorizeString("Fatal:" . $category . ":" . ToString($object), 'fatal'),
+      $where, $message, $detail_level, @details);
+    # If we're about to (really) DIE, we'll bypass the usual status message, so add it here.
+    # This really should be handled by the top-level program,
+    # after doing all processing within an eval
+    # BIZARRE: Note that die adds the "at <file> <line>" stuff IFF the message doesn't end w/ CR!
+    $message .= $state->getStatusMessage . "\n" if $state && !$ineval;
   }
   else {    # If we ARE in a recursive call, the actual message is $details[0]
     $message = $details[0] if $details[0]; }
-  local $LaTeXML::Common::Error::InHandler = 1;
+  # If inside an eval, this won't actually die, but WILL set $@ for caller's use.
   die $message; }
 
 sub checkRecursiveError {
@@ -74,32 +110,31 @@ sub checkRecursiveError {
       return 1; } }
   return; }
 
-# Note that "100" is hardwired into TeX, The Program!!!
-my $MAXERRORS = 100;    # [CONSTANT]
-
 # Should be fatal if strict is set, else warn.
 sub Error {
   my ($category, $object, $where, $message, @details) = @_;
-  my $state = $STATE;
+  my $state     = $STATE;
   my $verbosity = $state && $state->lookupValue('VERBOSITY') || 0;
-  if ($state->lookupValue('STRICT')) {
+  if ($state && $state->lookupValue('STRICT')) {
     Fatal($category, $object, $where, $message, @details); }
   else {
-    $state->noteStatus('error');
-    print STDERR generateMessage("Error:" . $category . ":" . ToString($object),
+    $state && $state->noteStatus('error');
+    print STDERR generateMessage(colorizeString("Error:" . $category . ":" . ToString($object), 'error'),
       $where, $message, 1, @details)
       if $verbosity >= -2; }
-  if (!$state || ($state->getStatus('error') || 0) > $MAXERRORS) {
-    Fatal('too_many_errors', $MAXERRORS, $where, "Too many errors (> $MAXERRORS)!"); }
+  # Note that "100" is hardwired into TeX, The Program!!!
+  my $maxerrors = ($state ? $state->lookupValue('MAX_ERRORS') : 100);
+  if ($state && (defined $maxerrors) && (($state->getStatus('error') || 0) > $maxerrors)) {
+    Fatal('too_many_errors', $maxerrors, $where, "Too many errors (> $maxerrors)!"); }
   return; }
 
 # Warning message; results may be OK, but somewhat unlikely
 sub Warn {
   my ($category, $object, $where, $message, @details) = @_;
-  my $state = $STATE;
+  my $state     = $STATE;
   my $verbosity = $state && $state->lookupValue('VERBOSITY') || 0;
   $state && $state->noteStatus('warning');
-  print STDERR generateMessage("Warning:" . $category . ":" . ToString($object),
+  print STDERR generateMessage(colorizeString("Warning:" . $category . ":" . ToString($object), 'warning'),
     $where, $message, 0, @details)
     if $verbosity >= -1;
   return; }
@@ -108,10 +143,10 @@ sub Warn {
 # but the message may give clues about subsequent warnings or errors
 sub Info {
   my ($category, $object, $where, $message, @details) = @_;
-  my $state = $STATE;
+  my $state     = $STATE;
   my $verbosity = $state && $state->lookupValue('VERBOSITY') || 0;
   $state && $state->noteStatus('info');
-  print STDERR generateMessage("Info:" . $category . ":" . ToString($object),
+  print STDERR generateMessage(colorizeString("Info:" . $category . ":" . ToString($object), 'info'),
     $where, $message, -1, @details)
     if $verbosity >= 0;
   return; }
@@ -163,24 +198,40 @@ sub NoteEnd {
 my $quoted_re     = qr/\"([^\"]*)\"/;                                                   # [CONSTANT]
 my $cantcall_re   = qr/Can't call method/;                                              # [CONSTANT]
 my $cantlocate_re = qr/Can't locate object method/;                                     # [CONSTANT]
+my $undef_re      = qr/Undefined subroutine/;                                           # [CONSTANT]
 my $noself_re     = qr/on an undefined value|without a package or object reference/;    # [CONSTANT]
 my $via_re        = qr/via package/;                                                    # [CONSTANT]
-my $at_re         = qr/at (.*)/;                                                        # [CONSTANT]
+my $at_re         = qr/(at .*)/;                                                        # [CONSTANT]
 
 sub perl_die_handler {
   my (@line) = @_;
+  if ($LaTeXML::IGNORE_ERRORS    # Just get out now, if we're ignoring errors within an eval.
+    || (colorstrip($line[0]) =~ /^\s*Fatal:/)) {    # Or, we've already been through here.
+    local $SIG{__DIE__} = undef;
+    die @line; }
   # We try to find a meaningful name for where the error occurred;
   # That's the thing that is "misdefined", after all.
   # Not completely sure we're looking in the right place up the stack, though.
-  if ($line[0] =~ /^$cantcall_re\s+$cantcall_re\s+($noself_re)\s+$at_re$/) {
+  if ($line[0] =~ /^$cantcall_re\s+$quoted_re\s+($noself_re)\s+$at_re$/) {
     my ($method, $kind, $where) = ($1, $2, $3);
-    Fatal('misdefined', callerName(2), $where, @line); }
-  elsif ($line[0] =~ /^$cantlocate_re\s+$quoted_re\s+$via_re\s+$quoted_re\s+$at_re$/) {
+    Fatal('misdefined', callerName(1), $where,
+      "Can't call method '$method' $kind", @line[1 .. $#line]); }
+  elsif ($line[0] =~ /^$undef_re\s+(\S+)\s+called $at_re$/) {
+    my ($function, $where) = ($1, $2);
+    Fatal('misdefined', callerName(1), $where,
+      "Undefined subroutine '$function' called", @line[1 .. $#line]); }
+  elsif ($line[0] =~ /^$cantlocate_re\s+$quoted_re\s+$via_re\s+$quoted_re\s+\(.*\)\s+$at_re/) {
     my ($method, $class, $where) = ($1, $2, $3);
-    Fatal('misdefined', callerName(2), $where, @line); }
-  elsif ($line[0] =~ /^Not an? (\w*) reference at (.*)$/) {
-    my ($type, $where) = ($1, $2);
-    Fatal('misdefined', callerName(2), $where, @line); }
+    Fatal('misdefined', callerName(1), $where,
+      "Can't locate method '$method' via '$class'", @line[1 .. $#line]); }
+  elsif ($line[0] =~ /^Can't locate \S* in \@INC \(you may need to install the (\S*) module\) \(\@INC contains: ([^\)]*)\) $at_re$/) {
+    my ($class, $inc, $where) = ($1, $2);
+    Fatal('misdefined', callerName(1), $where,
+      "Can't locate class '$class' (not installed or misspelled?)", @line[1 .. $#line]); }
+  elsif ($line[0] =~ /^Can't use\s+(\w*)\s+\([^\)]*\) as (.*?) ref(?:\s+while "strict refs" in use)? at (.*)$/) {
+    my ($gottype, $wanttype, $where) = ($1, $2, $3);
+    Fatal('misdefined', callerName(1), $where,
+      "Can't use $gottype as $wanttype reference", @line[1 .. $#line]); }
   elsif ($line[0] =~ /^File (.*?) had an error:/) {
     my ($file) = ($1);
     Fatal('misdefined', $file, undef, @line); }
@@ -190,9 +241,43 @@ sub perl_die_handler {
 
 sub perl_warn_handler {
   my (@line) = @_;
-  Warn('perl', 'warn', undef, "Perl warning", @line);
+  return if $LaTeXML::IGNORE_ERRORS;
+  if ($line[0] =~ /^Use of uninitialized value (.*?)(\s?+in .*?)\s+(at\s+.*?\s+line\s+\d+)\.$/) {
+    my ($what, $how, $where) = ($1 || 'value', $2, $3);
+    Warn('uninitialized', $what, $where, "Use of uninitialized value $what $how", @line[1 .. $#line]); }
+  elsif ($line[0] =~ /^Deep recursion on/) {
+    Fatal('perl', 'deep_recursion', undef, $line[0]); }
+  elsif ($line[0] =~ /^(.*?)\s+(at\s+.*?\s+line\s+\d+)\.$/) {
+    my ($warning, $where) = ($1, $2);
+    Warn('perl', 'warn', undef, $warning, $where, @line[1 .. $#line]); }
+  else {
+    Warn('perl', 'warn', undef, "Perl warning", @line); }
   return; }
 
+# The following handlers SHOULD report the problem,
+# even when within a "safe" eval that's ignoring errors.
+# Moreover, we'd really like to be able to throw all the way to
+# the top-level containing eval.  How to do that?
+sub perl_interrupt_handler {
+  my (@line) = @_;
+  $LaTeXML::IGNORE_ERRORS = 0;    # NOT ignored
+  $LaTeXML::UNSAFE_FATAL  = 1;
+  Fatal('interrupt', 'interrupted', undef, "LaTeXML was interrupted", @line);
+  return; }
+
+sub perl_timeout_handler {
+  my (@line) = @_;
+  $LaTeXML::IGNORE_ERRORS = 0;    # NOT ignored
+  $LaTeXML::UNSAFE_FATAL  = 1;
+  Fatal('timeout', 'timedout', undef, "Conversion timed out", @line);
+  return; }
+
+sub perl_terminate_handler {
+  my (@line) = @_;
+  $LaTeXML::IGNORE_ERRORS = 0;    # NOT ignored
+  $LaTeXML::UNSAFE_FATAL  = 1;
+  Fatal('terminate', 'terminated', undef, "Conversion was terminated", @line);
+  return; }
 #%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 # Internals
 # Synthesize an error message describing what happened, and where.
@@ -208,16 +293,20 @@ sub generateMessage {
   # Generate location information; basic and for stack trace.
   # If we've been given an object $where, where the error occurred, use it.
   my $docloc = getLocation($where);
+  $docloc = defined $docloc ? ToString($docloc) : '';
 
   # $message and each of @extra should be single lines
-  ($message, @extra) = grep { $_ ne '' } map { split("\n", $_) } grep { defined $_ } $message, @extra;
+  @extra = grep { $_ ne '' } map { split("\n", $_) } grep { defined $_ } $message, @extra;
+  # make 1st line be 1st line of message
+  $message = shift(@extra);
+  #  $message =~ s/\n.*//g;
   # The initial portion of the message will consist of:
   $message = '' unless defined $message;
   my @lines = (
     # Start with the error code & primary error message
     $errorcode . ' ' . $message,
     # Followed by single line location of where the message occurred (if we know)
-    ($docloc ? ($docloc) : ()),
+    ($docloc ? ('at ' . $docloc) : ()),
     # and then any additional message lines supplied
     @extra);
 
@@ -250,26 +339,51 @@ sub generateMessage {
 
   #----------------------------------------
   # Add Stack Trace, if that seems worthwhile.
-  if ($detail > -1) {
+  if (($detail > 1) && ($verbosity > 0)) {
+    push(@lines, "Stack Trace:", stacktrace()); }
+  elsif ($detail > -1) {
     my $nstack = ($detail > 1 ? undef : ($detail > 0 ? 4 : 1));
     if (my @objects = objectStack($nstack)) {
       my $top = shift(@objects);
-      push(@lines,   "In " . trim(ToString($top)) . ' ' . ToString(Locator($top)));
-      push(@objects, '...') if @objects && defined $nstack;
-      push(@lines,   join('', (map { ' <= ' . trim(ToString($_)) } @objects))) if @objects; } }
+      push(@lines, "In " . trim(Stringify($$top[0])) . ' ' . Stringify($$top[1]));
+      push(@objects, ['...']) if @objects && defined $nstack;
+      push(@lines, join('', (map { ' <= ' . trim(Stringify($$_[0])) } @objects))) if @objects;
+  } }
 
   # finally, join the result into a block of lines, indenting all but the 1st line.
   return "\n" . join("\n\t", @lines) . "\n"; }
 
+sub MergeStatus {
+  my ($external_state) = @_;
+  my $state = $STATE;
+  return unless $state && $external_state;
+  my $status          = $$state{status};
+  my $external_status = $$external_state{status};
+  # Should this be a state method? I suspect XS-ive conflicts later on...
+  foreach my $type (keys %$external_status) {
+    if ($type eq 'undefined' or $type eq 'missing') {
+      my $table = $$external_status{$type};
+      foreach my $subtype (keys %$table) {
+        $$status{$type}{$subtype} += $$table{$subtype};
+      }
+    } else {
+      $$status{$type} += $$external_status{$type};
+    }
+  }
+  return; }
+
+# returns the locator of an object, or undef
 sub Locator {
   my ($object) = @_;
-  return ($object && $object->can('getLocator') ? $object->getLocator : "???"); }
+  return ($object && $object->can('getLocator') ? $object->getLocator : undef); }
 
+# A more organized abstraction along there likes of $where->whereAreYou
+# might be useful?
 sub getLocation {
   my ($where) = @_;
   my $wheretype = ref $where;
   if ($wheretype && ($wheretype =~ /^XML::LibXML/)) {
-    my $box = $LaTeXML::DOCUMENT->getNodeBox($where);
+    my $box = $LaTeXML::DOCUMENT && $LaTeXML::DOCUMENT->getNodeBox($where);
     return Locator($box) if $box; }
   elsif ($wheretype && $where->can('getLocator')) {
     return $where->getLocator; }
@@ -287,7 +401,11 @@ sub getLocation {
     # NOTE: Problems here.
     # (1) With obsoleting Tokens as a Mouth, we can get pointless "Anonymous String" locators!
     # (2) If gullet is the source, we probably want to include next token, etc or
-    return $gullet->getLocator(); }
+    return Locator($gullet); }
+  # # If in postprocessing
+  # if($LaTeXML::Post::PROCESSOR && $LaTeXML::Post::DOCUMENT){
+  #   return 'in '. $LaTeXML::Post::PROCESSOR->getName
+  #     . ' on '. $LaTeXML::Post::DOCUMENT->siteRelativeDestination; }
   return; }
 
 sub callerName {
@@ -358,8 +476,8 @@ sub format_arg {
   elsif (ref $arg)         { $arg = Stringify($arg); }    # Allow overloaded stringify!
   elsif ($arg =~ /^-?[\d.]+\z/) { }                       # Leave numbers alone.
   else {                                                  # Otherwise, string, so quote
-    $arg =~ s/'/\\'/g;                                    # Slashify '
-    $arg =~ s/([[:cntrl::]])/ "\\".chr(ord($1)+ord('A'))/ge;
+    $arg =~ s/'/\\'/g;                                        # Slashify '
+    $arg =~ s/([[:cntrl:]])/ "\\".chr(ord($1)+ord('A'))/ge;
     $arg = "'$arg'" }
   return trim($arg); }
 
@@ -393,9 +511,11 @@ sub objectStack {
       my $method = $info{sub};
       $method =~ s/^.*:://;
       if ($self->can($method)) {
-        next if @objects && ($self eq $objects[-1]);
-        next unless $self->can('getLocator');
-        push(@objects, $self);
+        next if @objects && ($self eq $objects[-1][0]);    # but don't duplicate
+        if ($self->can('getLocator')) {                    # Digestion object?
+          push(@objects, [$self, Locator($self)]); }
+        elsif ($self->isa('LaTeXML::Post::Processor') || $self->isa('LaTeXML::Post::Document')) {
+          push(@objects, [$self, '->' . $method]); }
         last if $maxdepth && (scalar(@objects) >= $maxdepth); } } }
   return @objects; }
 
@@ -484,7 +604,7 @@ No user serviceable parts inside.  These symbols are not exported.
 
 Constructs an error or warning message based on the current stack and
 the current location in the document.
-C<$typ> is a short string characterizing the type of message, such as "Error".  
+C<$typ> is a short string characterizing the type of message, such as "Error".
 C<$msg> is the error message itself. If C<$lng> is true, will generate a
 more verbose message; this also uses the VERBOSITY set in the C<$STATE>.
 Longer messages will show a trace of the objects invoked on the stack,
@@ -513,4 +633,3 @@ Public domain software, produced as part of work done by the
 United States Government & not subject to copyright in the US.
 
 =cut
-
